/*
Copyright (C) 2006 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "DialogOpcodes.h"
#include "DebuggerCoreInterface.h"
#include "Debugger.h"
#include "MemoryRegions.h"
#include "ByteShiftArray.h"

#include <QHeaderView>
#include <QMessageBox>
#include <QVector>

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
DialogOpcodes::DialogOpcodes(QWidget *parent) : QDialog(parent) {
	ui.setupUi(this);
	ui.tableWidget->verticalHeader()->hide();
	ui.tableWidget->horizontalHeader()->setStretchLastSection(true);
}


//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogOpcodes::showEvent(QShowEvent *) {
	if(Debugger::memoryRegions != 0) {
		Debugger::memoryRegions->sync();
		QVector<MemoryRegions::Region> r = Debugger::memoryRegions->regions();

		ui.tableWidget->setSortingEnabled(false);
		ui.tableWidget->setRowCount(0);
		for(QVector<MemoryRegions::Region>::const_iterator i = r.begin(); i != r.end(); ++i) {
			int row = ui.tableWidget->rowCount();
			ui.tableWidget->insertRow(row);
			ui.tableWidget->setItem(row, 0, new QTableWidgetItem(QString().sprintf("%08x", i->start)));
			ui.tableWidget->setItem(row, 1, new QTableWidgetItem(QString().sprintf("%08x", i->end)));
			ui.tableWidget->setItem(row, 2, new QTableWidgetItem(QString().sprintf("%c%c%c", i->readable() ? 'r' : '-', i->writeable() ? 'w' : '-', i->executable() ? 'x' : '-' )));
			ui.tableWidget->setItem(row, 3, new QTableWidgetItem(i->name));
		}
		ui.tableWidget->setSortingEnabled(true);
	} else {
		ui.tableWidget->setRowCount(0);
	}
	ui.progressBar->setValue(0);
	ui.listWidget->clear();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogOpcodes::testEax(const DialogOpcodes::OpcodeData &data, BaseTypes::address_t start_address) {
	switch(data.word) {
	case 0xe0ff: ui.listWidget->addItem(QString().sprintf("%08x: jump eax", start_address)); break;
	case 0xd0ff: ui.listWidget->addItem(QString().sprintf("%08x: call eax", start_address)); break;
	case 0xc350: ui.listWidget->addItem(QString().sprintf("%08x: push eax; ret", start_address)); break;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogOpcodes::testEbx(const DialogOpcodes::OpcodeData &data, BaseTypes::address_t start_address) {
	switch(data.word) {
	case 0xe3ff: ui.listWidget->addItem(QString().sprintf("%08x: jump ebx", start_address)); break;
	case 0xd3ff: ui.listWidget->addItem(QString().sprintf("%08x: call ebx", start_address)); break;
	case 0xc353: ui.listWidget->addItem(QString().sprintf("%08x: push ebx; ret", start_address)); break;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogOpcodes::testEcx(const DialogOpcodes::OpcodeData &data, BaseTypes::address_t start_address) {
	switch(data.word) {
	case 0xe1ff: ui.listWidget->addItem(QString().sprintf("%08x: jump ecx", start_address)); break;
	case 0xd1ff: ui.listWidget->addItem(QString().sprintf("%08x: call ecx", start_address)); break;
	case 0xc351: ui.listWidget->addItem(QString().sprintf("%08x: push ecx; ret", start_address)); break;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogOpcodes::testEdx(const DialogOpcodes::OpcodeData &data, BaseTypes::address_t start_address) {
	switch(data.word) {
	case 0xe2ff: ui.listWidget->addItem(QString().sprintf("%08x: jump edx", start_address)); break;
	case 0xd2ff: ui.listWidget->addItem(QString().sprintf("%08x: call edx", start_address)); break;
	case 0xc352: ui.listWidget->addItem(QString().sprintf("%08x: push edx; ret", start_address)); break;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogOpcodes::testEbp(const DialogOpcodes::OpcodeData &data, BaseTypes::address_t start_address) {
	switch(data.word) {
	case 0xe5ff: ui.listWidget->addItem(QString().sprintf("%08x: jump ebp", start_address)); break;
	case 0xd5ff: ui.listWidget->addItem(QString().sprintf("%08x: call ebp", start_address)); break;
	case 0xc355: ui.listWidget->addItem(QString().sprintf("%08x: push ebp; ret", start_address)); break;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogOpcodes::testEsp(const DialogOpcodes::OpcodeData &data, BaseTypes::address_t start_address) {
	switch(data.word) {
	case 0xe4ff: ui.listWidget->addItem(QString().sprintf("%08x: jump esp", start_address)); break;
	case 0xd4ff: ui.listWidget->addItem(QString().sprintf("%08x: call esp", start_address)); break;
	case 0xc354: ui.listWidget->addItem(QString().sprintf("%08x: push esp; ret", start_address)); break;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogOpcodes::testEsi(const DialogOpcodes::OpcodeData &data, BaseTypes::address_t start_address) {
	switch(data.word) {
	case 0xe6ff: ui.listWidget->addItem(QString().sprintf("%08x: jump esi", start_address)); break;
	case 0xd6ff: ui.listWidget->addItem(QString().sprintf("%08x: call esi", start_address)); break;
	case 0xc356: ui.listWidget->addItem(QString().sprintf("%08x: push esi; ret", start_address)); break;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogOpcodes::testEdi(const DialogOpcodes::OpcodeData &data, BaseTypes::address_t start_address) {
	switch(data.word) {
	case 0xe7ff: ui.listWidget->addItem(QString().sprintf("%08x: jump edi", start_address)); break;
	case 0xd7ff: ui.listWidget->addItem(QString().sprintf("%08x: call edi", start_address)); break;
	case 0xc357: ui.listWidget->addItem(QString().sprintf("%08x: push edi; ret", start_address)); break;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogOpcodes::testESP_ADD_4(const OpcodeData &data, BaseTypes::address_t start_address) {
	switch(data.word) {
	case 0xc358: ui.listWidget->addItem(QString().sprintf("%08x: pop eax; ret", start_address)); break;
	case 0xc35b: ui.listWidget->addItem(QString().sprintf("%08x: pop ebx; ret", start_address)); break;
	case 0xc359: ui.listWidget->addItem(QString().sprintf("%08x: pop ecx; ret", start_address)); break;
	case 0xc35a: ui.listWidget->addItem(QString().sprintf("%08x: pop edx; ret", start_address)); break;
	case 0xc35e: ui.listWidget->addItem(QString().sprintf("%08x: pop esi; ret", start_address)); break;
	case 0xc35f: ui.listWidget->addItem(QString().sprintf("%08x: pop edi; ret", start_address)); break;
	case 0xc35d: ui.listWidget->addItem(QString().sprintf("%08x: pop ebp; ret", start_address)); break;
	case 0xc35c: ui.listWidget->addItem(QString().sprintf("%08x: pop esp; ret", start_address)); break;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogOpcodes::testESP_SUB_4(const OpcodeData &data, BaseTypes::address_t start_address) {
	switch(data.dword) {
	case 0xfc2464ff: ui.listWidget->addItem(QString().sprintf("%08x: jmp [esp - 4]", start_address)); break;
	case 0xfc2454ff: ui.listWidget->addItem(QString().sprintf("%08x: call [esp - 4]", start_address)); break;
	case 0xc304ec83: ui.listWidget->addItem(QString().sprintf("%08x: sub esp, 4; ret", start_address)); break;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogOpcodes::testESP_ADD_0(const OpcodeData &data, BaseTypes::address_t start_address) {
	switch(data.byte) {
	case 0xc3: ui.listWidget->addItem(QString().sprintf("%08x: ret", start_address)); break;
	case 0xc2: ui.listWidget->addItem(QString().sprintf("%08x: ret N", start_address)); break;
	}
	
	if(std::memcmp(data.data, "\x58\xff\xd0", 3) == 0) ui.listWidget->addItem(QString().sprintf("%08x: pop eax; call eax", start_address));
	if(std::memcmp(data.data, "\x59\xff\xd1", 3) == 0) ui.listWidget->addItem(QString().sprintf("%08x: pop ecx; call ecx", start_address));
	if(std::memcmp(data.data, "\x5a\xff\xd2", 3) == 0) ui.listWidget->addItem(QString().sprintf("%08x: pop edx; call edx", start_address));
	if(std::memcmp(data.data, "\x5b\xff\xd3", 3) == 0) ui.listWidget->addItem(QString().sprintf("%08x: pop ebx; call ebx", start_address));
	if(std::memcmp(data.data, "\x5c\xff\xd4", 3) == 0) ui.listWidget->addItem(QString().sprintf("%08x: pop esp; call esp", start_address));
	if(std::memcmp(data.data, "\x5d\xff\xd5", 3) == 0) ui.listWidget->addItem(QString().sprintf("%08x: pop ebp; call ebp", start_address));
	if(std::memcmp(data.data, "\x5e\xff\xd6", 3) == 0) ui.listWidget->addItem(QString().sprintf("%08x: pop esi; call esi", start_address));
	if(std::memcmp(data.data, "\x5f\xff\xd7", 3) == 0) ui.listWidget->addItem(QString().sprintf("%08x: pop edi; call edi", start_address));
	if(std::memcmp(data.data, "\x58\xff\xe0", 3) == 0) ui.listWidget->addItem(QString().sprintf("%08x: pop eax; jmp eax", start_address));
	if(std::memcmp(data.data, "\x59\xff\xe1", 3) == 0) ui.listWidget->addItem(QString().sprintf("%08x: pop ecx; jmp ecx", start_address));
	if(std::memcmp(data.data, "\x5a\xff\xe2", 3) == 0) ui.listWidget->addItem(QString().sprintf("%08x: pop edx; jmp edx", start_address));
	if(std::memcmp(data.data, "\x5b\xff\xe3", 3) == 0) ui.listWidget->addItem(QString().sprintf("%08x: pop ebx; jmp ebx", start_address));
	if(std::memcmp(data.data, "\x5c\xff\xe4", 3) == 0) ui.listWidget->addItem(QString().sprintf("%08x: pop esp; jmp esp", start_address));
	if(std::memcmp(data.data, "\x5d\xff\xe5", 3) == 0) ui.listWidget->addItem(QString().sprintf("%08x: pop ebp; jmp ebp", start_address));
	if(std::memcmp(data.data, "\x5e\xff\xe6", 3) == 0) ui.listWidget->addItem(QString().sprintf("%08x: pop esi; jmp esi", start_address));
	if(std::memcmp(data.data, "\x5f\xff\xe7", 3) == 0) ui.listWidget->addItem(QString().sprintf("%08x: pop edi; jmp edi", start_address));
	if(std::memcmp(data.data, "\xff\x24\x24", 3) == 0) ui.listWidget->addItem(QString().sprintf("%08x: jmp [esp]", start_address));
	if(std::memcmp(data.data, "\xff\x14\x24", 3) == 0) ui.listWidget->addItem(QString().sprintf("%08x: call [esp]", start_address));
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogOpcodes::on_btnFind_clicked() {

	QList<QTableWidgetItem *> sel = ui.tableWidget->selectedItems();
	
	ByteShiftArray bsa(sizeof(OpcodeData));
	
	int classtype = ui.comboBox->currentIndex();

	ui.listWidget->clear();
	ui.progressBar->setValue(0);
	
	if(sel.size() == 0) {
		QMessageBox::information(this, "No Region Selected", "You must select a region which is to be scanned for the desired opcode.");
	} else {	
		
		BaseTypes::address_t start_address	= sel[0]->text().toUInt(0, 16);
		BaseTypes::address_t end_address	= sel[1]->text().toUInt(0, 16);

		const BaseTypes::address_t orig_start	= start_address;
		const BaseTypes::address_t howMany		= end_address - orig_start;
		bsa.clear();
		
		// create a reference to the bsa's data so we can pass it to the testXXXX functions
		const OpcodeData &opcode = *reinterpret_cast<const OpcodeData *>(bsa.data());
		
		// do the search for this region!
		while(start_address < end_address) {

			// read in the next byte
			bsa.rol();
			Debugger::debuggerBase->readBytes(start_address, &(bsa[bsa.size() - 1]), sizeof(uint8_t));

			switch(classtype) {
			case 0: 
				testEax(opcode, start_address - 3);					
				break;
			case 1:
				testEbx(opcode, start_address - 3);
				break;
			case 2:
				testEcx(opcode, start_address - 3);
				break;
			case 3:
				testEdx(opcode, start_address - 3);
				break;
			case 4:
				testEbp(opcode, start_address - 3);
				break;
			case 5:
				testEsp(opcode, start_address - 3);
				break;
			case 6:
				testEsi(opcode, start_address - 3);
				break;
			case 7:
				testEdi(opcode, start_address - 3);
				break;
			case 8:
				testEax(opcode, start_address - 3);					
				testEbx(opcode, start_address - 3);
				testEcx(opcode, start_address - 3);
				testEdx(opcode, start_address - 3);
				testEbp(opcode, start_address - 3);
				testEsp(opcode, start_address - 3);
				testEsi(opcode, start_address - 3);
				testEdi(opcode, start_address - 3);
				break;
			case 9:
				// [ESP] -> EIP
				testESP_ADD_0(opcode, start_address - 3);
				break;	
			case 10:
				// [ESP + 4] -> EIP
				testESP_ADD_4(opcode, start_address - 3);
				break;
			case 11:
				// [ESP - 4] -> EIP
				testESP_SUB_4(opcode, start_address - 3);
				break;
			}

			const BaseTypes::address_t howManyDone	= start_address - orig_start;
			const int percent = static_cast<int>(static_cast<double>(howManyDone) / static_cast<double>(howMany) * 100);			

			ui.progressBar->setValue(percent);
			++start_address;
		}
		ui.progressBar->setValue(100);
	}
}
