#!/usr/bin/perl

my %tables;
my @table_table;
my ($i, $name, $current_table, $entries);
my ($insn, $line, $list, @items);
my ($byte, $cpu, $isa, $mflist, $dflist, $sflist, $aftlist, $eflags);
my $last_line = "";
my $pfx="ia32";
my @lines;


$i = 0;

open(FILE, shift());

foreach (<FILE>){
	chomp;
	if ( /^#/ or /^\s*$/ ) {
		next;
	} elsif ( /^PREFIX ([A-Za-z0-9_]+)/ ) {
		# set prefix for variable names
		$pfx = $1;
	} elsif ( /^TABLE ([A-Za-z0-9_]+)\t(tbl_[a-z_]+)\tshift ([0-9]+)\tmask ([0-9]+)\tminlim ([0-9]+)\tmaxlim ([0-9]+)\t"([^"]*)"$/ ) {
		# add to @tables

		$table_table[$i] = {};
		$current_table = $table_table[$i];
		$$current_table{'name'} = $pfx . '_' . $1;
		$$current_table{'type'} = table_type($2);
		$$current_table{'index'} = $i; #= $3;
		$$current_table{'shift'} = $3;
		$$current_table{'mask'} = $4;
		$$current_table{'min'} = $5;
		$$current_table{'max'} = $6;
		$$current_table{'descr'} = $7;
		$$current_table{'prefix'} = "pfx_" . $1;
		$$current_table{'prefix_entries'} = [];
		$$current_table{'insn_list'} = [];
		$$current_table{'next_insn_byte'} = $5;

		$i = $i + 1;
		
	} elsif ( /^INSN (.+)$/ ) {
		
		$insn = $1;
		$insn =~ s/ //g;

		#Table|MnemFlag|DestFlag|SrcFlag|AuxFlag|CPU|mnem|dest|src|aux|flags_effected|ipmlicit|cmt
		($t,$mf,$df,$sf,$af,$cpuisa,$m,$d,$s,$a,$flg,$imp,$cmt) = 
			split '\t', $insn;

		# -------------------------------------------------
		# OK, add some custom code here to modify the insn
		# -------------------------------------------------

		$byte = $$current_table{'next_insn_byte'};

		$m =~ s/"//g;
		$t = table($t);
		$info = info($mf, $cpuisa);
		$dest = operand($d, $df);
		$src = operand($s, $sf);
		$aux = operand($a, $af);
		$imp = implicit($imp);
		$eflags = eflags($flg);

		$line = "{ 'byte':$byte, 'table':$t," . 
			 " 'info':$info, 'mnemonic':'$m'," .
			 " 'dest':$dest, 'src':$src, 'imm':$aux," .
			 " 'implicit':$imp, 'eflags':$eflags }";

		# add to $tables[$#tables]{lines}
		push @{$$current_table{'insn_list'}}, $line;
		$$current_table{'next_insn_byte'} = $byte + 1;

	} elsif ( /^END TABLE/ ) {
		# no need to do anything
	}
}
close(FILE);

# output opcode tables
print "#______________________________________________________________\n";
print "# Opcode tables: \n";
$i = 0;
foreach ( @table_table ) {
	print "# TABLE " . $$_{'name'} . " ___________________________\n";
	print $$_{'name'} . " = (\n";

	$list = \@{$$_{'insn_list'}};
	foreach ( @{$list} ) {
		print "\n$_";
		if ( $i != $#$list ) {
			print ',';
		}
		print "\n";
		$i = $i + 1;
	}
	print ")\n\n\n";
}

# output table of opcode tables
print "#______________________________________________________________\n";
print "# Table of opcode tables: \n";
print "${pfx}_tables = (\n";
$i = 0;
foreach ( @table_table ) {
	printf "\t{ 'index':%d, 'table':%s, ", $$_{'index'}, $$_{'name'};
	printf "'name':'%s', 'type':'%s', ", $$_{'name'} , $$_{'type'};
	printf "'shift':0x%02X, 'mask':0x%02X, ", $$_{'shift'}, $$_{'mask'}; 
	printf "'min':0x%02X, 'max':0x%02X }", $$_{'min'}, $$_{'max'};

	if ( $i != $#table_table ) {
		print ",";

	}

	print "\n";
	$i = $i + 1;
}
print ")\n\n";

sub table() {

	my $tbl = shift;
	my %table_indexes = ( 'idx_Main' => 0, 'idx_66' => 1, 'idx_F2' => 2,
	'idx_F3' => 3, 'idx_0F' => 4, 'idx_660F' => 5, 'idx_F20F' => 6,
	'idx_F30F' => 7, 'idx_0F00' => 8, 'idx_0F01' => 9, 'idx_0F0111' => 10,
	'idx_0F12' => 11, 'idx_0F16' => 12, 'idx_0F18' => 13, 'idx_0F71' => 14,
	'idx_660F71' => 15, 'idx_0F72' => 16, 'idx_660F72' => 17,
	'idx_0F73' => 18, 'idx_660F73' => 19, 'idx_0FAE' => 20,
	'idx_0FBA' => 21, 'idx_0FC7' => 22, 'idx_0FB9' => 23,
	'idx_C6' => 24, 'idx_C7' => 25, 'idx_80' => 26, 'idx_81' => 27, 
	'idx_82' => 28, 'idx_83' => 29, 'idx_C0' => 30, 'idx_C1' => 31, 
	'idx_D0' => 32, 'idx_D1' => 33, 'idx_D2' => 34, 'idx_D3' => 35,
	'idx_F6' => 36, 'idx_F7' => 37, 'idx_FE' => 38, 'idx_FF' => 39,
	'idx_D8' => 40, 'idx_D8C0' => 41, 'idx_D9' => 42, 'idx_D9C0' => 43,
	'idx_DA' => 44, 'idx_DAC0' => 45, 'idx_DB' => 46, 'idx_DBC0' => 47,
	'idx_DC' => 48, 'idx_DCC0' => 49, 'idx_DD' => 50, 'idx_DDC0' => 51,
	'idx_DE' => 52, 'idx_DEC0' => 53, 'idx_DF' => 54, 'idx_DFC0' => 55,
	'idx_0F0F' => 56 );


	$tbl =~ s/\s//g;
	if ( $tbl =~ /idx_/ ) {
		$tbl = $table_indexes{$tbl};
	}

	return $tbl;
	my $line = "";
}

sub table_type() {
	my $type = shift;
	$type =~ s/^tbl_//;
	return $type;
}

sub eflags() {
	my ($flags) = @_;
	my $line = "";
	my $tmp;
	my @set = (), @test = ();

	if ( $flags eq '0' ) {
		return "{'set':'', 'test':''}";
	}

	foreach (split /\|/, $flags) {
		$tmp = $_;
		$tmp =~ s/INS_//;
		if ( $tmp =~ /SET_ALL/ ) {
			push @set, 'carry','zero','overflow','sign','parity';
		} elsif ( $tmp =~ /SET_(.*)/ ) {
			push @set, lc $1;
		} elsif ( $tmp =~ /TEST_ALL/ ) {
			push @test, 'carry','zero','overflow','sign','parity';
		} elsif ( $tmp =~ /TEST_(.*)/ ) {
			push @test, lc $1;
		}
		$line = $line . "'$tmp', ";
	}

	$line = "'set':'";
	foreach (@set) {
		$line .= "$_ ";
	}
	$line =~ s/\s$//;

	$line .= "', 'test':'";
	foreach (@test) {
		$line .= "$_ ";
	}
	$line =~ s/\s$//;
	$line .= "'";

	return "{$line}";
}


sub info () {
	my ($flags, $cpuisa) = @_;
	my $note_prefix = "False";
	my $note_suffix = "False";
	my $note_ring0 = "False";
	my $note_smm = "False";
	my $note_serial = "False";
	my $prefix = "";
	my $info = "";
	my (@items, $cpu, $isa, $notes);

	my %insn_types = (
	 'INS_BRANCH' => "'type':'branch', 'group':'exec'",
	 'INS_BRANCHCC' => "'type':'branchcc', 'group':'exec'",
	 'INS_CALL' => "'type':'call', 'group':'exec'",
	 'INS_CALLCC' => "'type':'callcc', 'group':'exec'",
	 'INS_RET' => "'type':'ret', 'group':'exec'",
	 'INS_ADD' => "'type':'add', 'group':'arith'",
	 'INS_SUB' => "'type':'sub', 'group':'arith'",
	 'INS_MUL' => "'type':'mul', 'group':'arith'",
	 'INS_DIV' => "'type':'div', 'group':'arith'",
	 'INS_INC' => "'type':'inc', 'group':'arith'",
	 'INS_DEC' => "'type':'dec', 'group':'arith'",
	 'INS_SHL' => "'type':'shl', 'group':'arith'",
	 'INS_SHR' => "'type':'shr', 'group':'arith'",
	 'INS_ROL' => "'type':'rol', 'group':'arith'",
	 'INS_ROR' => "'type':'ror', 'group':'arith'",
	 'INS_MIN' => "'type':'min', 'group':'arith'",
	 'INS_MAX' => "'type':'max', 'group':'arith'",
	 'INS_AVG' => "'type':'avg', 'group':'arith'",
	 'INS_FLR' => "'type':'flr', 'group':'arith'",
	 'INS_CEIL' => "'type':'ceil', 'group':'arith'",
	 'INS_AND' => "'type':'and', 'group':'logic'",
	 'INS_OR' => "'type':'or', 'group':'logic'",
	 'INS_XOR' => "'type':'xor', 'group':'logic'",
	 'INS_NOT' => "'type':'not', 'group':'logic'",
	 'INS_NEG' => "'type':'neg', 'group':'logic'",
	 'INS_NAND' => "'type':'nand', 'group':'logic'",
	 'INS_PUSH' => "'type':'push', 'group':'stack'",
	 'INS_POP' => "'type':'pop', 'group':'stack'",
	 'INS_PUSHREGS' => "'type':'pushregs', 'group':'stack'",
	 'INS_POPREGS' => "'type':'popregs', 'group':'stack'",
	 'INS_PUSHFLAGS' => "'type':'pushflags', 'group':'stack'",
	 'INS_POPFLAGS' => "'type':'popflags', 'group':'stack'",
	 'INS_ENTER' => "'type':'enter', 'group':'stacK'",
	 'INS_LEAVE' => "'type':'leave', 'group':'stack'",
	 'INS_TEST' => "'type':'test', 'group':'cond'",
	 'INS_CMP' => "'type':'cmp', 'group':'cond'",
	 'INS_MOV' => "'type':'mov', 'group':'load'",
	 'INS_MOVCC' => "'type':'movcc', 'group':'load'",
	 'INS_XCHG' => "'type':'xchg', 'group':'load'",
	 'INS_XCHGCC' => "'type':'xchgcc', 'group':'load'",
	 'INS_CONV' => "'type':'conv', 'group':'load'",
	 'INS_STRCMP' => "'type':'strcmp', 'group':'array'",
	 'INS_STRLOAD' => "'type':'strload', 'group':'array'",
	 'INS_STRMOV' => "'type':'strmov', 'group':'array'",
	 'INS_STRSTOR' => "'type':'strstor', 'group':'array'",
	 'INS_XLAT' => "'type':'xlat', 'group':'array'",
	 'INS_BITTEST' => "'type':'bittest', 'group':'bit'",
	 'INS_BITSET' => "'type':'bitset', 'group':'bit'",
	 'INS_BITCLR' => "'type':'bitclr', 'group':'bit'",
	 'INS_CLEARCF' => "'type':'clearcf', 'group':'flag'",
	 'INS_CLEARZF' => "'type':'clearzf', 'group':'flag'",
	 'INS_CLEAROF' => "'type':'clearof', 'group':'flag'",
	 'INS_CLEARDF' => "'type':'cleardf', 'group':'flag'",
	 'INS_CLEARSF' => "'type':'clearsf', 'group':'flag'",
	 'INS_CLEARPF' => "'type':'clearpf', 'group':'flag'",
	 'INS_SETCF' => "'type':'setcf', 'group':'flag'",
	 'INS_SETZF' => "'type':'setzf', 'group':'flag'",
	 'INS_SETOF' => "'type':'setof', 'group':'flag'",
	 'INS_SETDF' => "'type':'setdf', 'group':'flag'",
	 'INS_SETSF' => "'type':'setsf', 'group':'flag'",
	 'INS_SETPF' => "'type':'setpf', 'group':'flag'",
	 'INS_TOGCF' => "'type':'togcf', 'group':'flag'",
	 'INS_TOGZF' => "'type':'togzf', 'group':'flag'",
	 'INS_TOGOF' => "'type':'togof', 'group':'flag'",
	 'INS_TOGDF' => "'type':'togdf', 'group':'flag'",
	 'INS_TOGSF' => "'type':'togsf', 'group':'flag'",
	 'INS_TOGPF' => "'type':'togpf', 'group':'flag'",
	 'INS_FMOV' => "'type':'fmov', 'group':'fpu'",
	 'INS_FMOVCC' => "'type':'fmovcc', 'group':'fpu'",
	 'INS_FNEG' => "'type':'fneg', 'group':'fpu'",
	 'INS_FABS' => "'type':'fabs', 'group':'fpu'",
	 'INS_FADD' => "'type':'fadd', 'group':'fpu'",
	 'INS_FSUB' => "'type':'fsub', 'group':'fpu'",
	 'INS_FMUL' => "'type':'fmul', 'group':'fpu'",
	 'INS_FDIV' => "'type':'fdiv', 'group':'fpu'",
	 'INS_FSQRT' => "'type':'fsqrt', 'group':'fpu'",
	 'INS_FCMP' => "'type':'fcmp', 'group':'fpu'",
	 'INS_FCOS' => "'type':'fcos', 'group':'fpu'",
	 'INS_FLDPI' => "'type':'fldpi', 'group':'fpu'",
	 'INS_FLDZ' => "'type':'fldz', 'group':'fpu'",
	 'INS_FTAN' => "'type':'ftan', 'group':'fpu'",
	 'INS_FSINE' => "'type':'fsine', 'group':'fpu'",
	 'INS_FSYS' => "'type':'fsys', 'group':'fpu'",
	 'INS_TRAP' => "'type':'trap', 'group':'traps'",
	 'INS_TRAPCC' => "'type':'trapcc', 'group':'traps'",
	 'INS_TRET' => "'type':'tret', 'group':'traps'",
	 'INS_BOUNDS' => "'type':'bounds', 'group':'traps'",
	 'INS_DEBUG' => "'type':'debug', 'group':'traps'",
	 'INS_TRACE' => "'type':'trace', 'group':'traps'",
	 'INS_INVALIDOP' => "'type':'invalidop', 'group':'traps'",
	 'INS_OFLOW' => "'type':'oflow', 'group':'traps'",
	 'INS_HALT' => "'type':'halt', 'group':'system'",
	 'INS_IN' => "'type':'in', 'group':'system'",
	 'INS_OUT' => "'type':'out', 'group':'system'",
	 'INS_CPUID' => "'type':'cpuid', 'group':'system'",
	 'INS_NOP' => "'type':'nop', 'group':'other'",
	 'INS_BCDCONV' => "'type':'bcdconv', 'group':'other'",
	 'INS_SZCONV' => "'type':'szconv', 'group':'other'",
	 'INS_UNKNOWN' => "'type':'unknown', 'group':'other'",
	 'INS_EXEC' => "'type':'unknown', 'group':'exec'",
	 'INS_ARITH' => "'type':'unknown', 'group':'arith'",
	 'INS_LOGIC' => "'type':'unknown', 'group':'logic'",
	 'INS_STACK' => "'type':'unknown', 'group':'stack'",
	 'INS_COND' => "'type':'unknown', 'group':'cond'",
	 'INS_LOAD' => "'type':'unknown', 'group':'load'",
	 'INS_ARRAY' => "'type':'unknown', 'group':'array'",
	 'INS_BIT' => "'type':'unknown', 'group':'bit'",
	 'INS_FLAG' => "'type':'unknown', 'group':'flag'",
	 'INS_FPU' => "'type':'unknown', 'group':'fpu'",
	 'INS_TRAPS' => "'type':'unknown', 'group':'traps'",
	 'INS_SYSTEM' => "'type':'unknown', 'group':'system'",
	 'INS_OTHER' => "'type':'unknown', 'group':'other'",
	 'INS_INVALID' => "'type':'invalid', 'group':'invalid'"
	);
	my %prefix_types = (
	"PREFIX_LOCK" => "'prefix_name':'lock', 'prefix_group':1",
	"PREFIX_REPNZ" => "'prefix_name':'repnz', 'prefix_group':1",
	"PREFIX_REPZ" => "'prefix_name':'repz', 'prefix_group':1",
	"PREFIX_OP_SIZE" => "'prefix_name':'op_size', 'prefix_group':3",
	"PREFIX_ADDR_SIZE" => "'prefix_name':'addr_size', 'prefix_group':4",
	"PREFIX_CS" => "'prefix_name':'cs', 'prefix_group':2",
	"PREFIX_SS" => "'prefix_name':'ss', 'prefix_group':2",
	"PREFIX_DS" => "'prefix_name':'ds', 'prefix_group':2",
	"PREFIX_ES" => "'prefix_name':'es', 'prefix_group':2",
	"PREFIX_FS" => "'prefix_name':'fs', 'prefix_group':2",
	"PREFIX_GS" => "'prefix_name':'gs', 'prefix_group':2",
	"PREFIX_TAKEN" => "'prefix_name':'taken', 'prefix_group':2",
	"PREFIX_NOTTAKEN" => "'prefix_name':'nottaken', 'prefix_group':2"
	);

	if ($cpuisa eq '0' ) {
		$cpu = 'None';
		$isa = 'None';
	} else {
		@items = split /\|/, $cpuisa;
		if ($#items) {
			$cpu = lc $items[0];
			$isa = lc $items[1];
			$isa =~ s/^isa_//;
		} else {
			$cpu = lc $cpuisa;
			$isa = "gp";
		}
		$cpu =~ s/^cpu_//;
	}

	$isa =~ s/gp/general/;

	foreach (split /\|/, $flags) {
		$_ =~ s/ //g;
		if ( /INS_NOTE_/ ) {
			# a bit silly, but who cares ;)
			if ( /PREFIX/ ) {
				$note_prefix = "True";
			} elsif ( /SUFFIX/ ) {
				$note_suffix = "True";
				$info = "'type':'subtable', 'group':'other'";
			} elsif ( /RING0/ ) {
				$note_ring0 = "True";
			} elsif ( /SMM/ ) {
				$note_smm = "True";
			} elsif ( /SERIAL/ ) {
				$note_serial = "True";
			}
			# this is a flag
		} elsif ( /PREFIX/ ) {
			# this is a prefix
			$prefix = ", " . $prefix_types{$_};
			$info = "'type':'prefix', 'group':'other'";
		} elsif ( /INS_/ ) {
			# this is an instruction type code
			$info = $insn_types{$_};
		} elsif ( /0/ ) {
			# this is an escape to another opcode table
			$info = "'type':'subtable', 'group':'other'";
		}
	}

	$notes = "'prefix':$note_prefix, 'suffix':$note_suffix, " . 
	         "'ring0':$note_ring0, 'smm':$note_smm, " .
	         "'serializing':$note_serial, 'cpu':'$cpu', 'isa':'$isa'" . 
		 $prefix;

	if ($info eq "" ) {
		$info = $notes;
	} else {
		$info = "$info, $notes";
	}

	return "{$info}";
}

sub operand() {
	my ($val, $flags) = @_;
	my $addrmeth = "";
	my $optype = "";
	my $signed = 'False';
	my %perms = ( 'R' => '-', 'W' => '-', 'X' => '-' );	
	my ($op, $tmp);

	if ( $flags =~ /^ARG_NONE/ ) {
		return 'None';
	}

	$op = "'value':$val,";

	foreach (split /\|/, $flags) {
		$_ =~ s/ //g;
		if ( /ADDRMETH_/ ) {
			$tmp = $_;
			$tmp =~ s/ADDRMETH_//g;
			$addrmeth = "'addr_method':'$tmp', ";
		} elsif ( /OPTYPE_/ ) {
			$tmp = $_;
			$tmp =~ s/OPTYPE_//g;
			$optype = "'datatype':'$tmp', ";
		} elsif ( /OP_SIGNED/ ) {
			$signed = 'True';
		} elsif ( /^OP_([RWX])/ ) {
			$perms{$1} = lc($1);
		} else {
			print "ERROR: $_\n";
		}
	}

	$tmp = $perms{'R'} . $perms{'W'} . $perms{'X'};
	$op = $op . $addrmeth . $optype . "'signed':$signed, 'access':'$tmp'";

	return "{$op}";
}

sub implicit() {
	my $imp = shift;
	if ($imp eq '0' or $imp eq "" ) {
		return 'None';
	}
	return $imp;
}

