#!/usr/bin/env python2.4
'''
	elf_disasm : hacked-together ELF disassembler based on BGO 
	             file loader test code for ELF files.
	WARNING: this is for crude testing ONLY and should not be
	         used for actual ELF disassembly!
'''

import sys
import struct
import re

sys.path.append(".")

import libdisasm.objfile as ObjFile
import libdisasm.disasm as libdisasm
# syntax types
import libdisasm.syntax.raw as Raw
import libdisasm.syntax.native as Native
import libdisasm.syntax.intel as Intel
import libdisasm.syntax.att as ATT
# output formats
import libdisasm.output.ascii as Ascii

class EIdent(object):
	_format = "<4sBBBBB7s"
	file_class_strings = ( 'none', '32', '64' )
	data_strings = ( 'None', 'LSB (little endian)', 'MSB (big endian)' )
	os_abi_strings = { 0:'SysV', 1:'HPUX', 2:'NetBSD', 3:'Linux',
	                   6:'Solaris', 7:'AIX', 8:'Irix', 9:'FreeBSD', 
	                   10:'Tru64', 11:'Modesto', 12:'OpenBSD', 97:'ARM',
	                   255:'Standalone' }
	def __init__(self, file, offset):
		handle = file.handle()
		size = struct.calcsize(self._format)
		try:
			handle.seek(offset)
			buf = handle.read(size)
		except Exception, e:
			sys.stderr.write( str(e) )
			raise e
		
		try:
			ident = struct.unpack(self._format, buf)
		except Exception, e:
			sys.stderr.write( str(e) )
			raise e
		
		self.magic = ident[0]
		
		self.file_class_raw = ident[1]
		try:
			self.file_class = self.file_class_strings[ident[1]]
		except IndexError, e:
			self.file_class = 'Invalid file class: ' + str(ident[1])
		
		self.data_raw = ident[2]
		try:
			self.data = self.data_strings[ident[2]]
		except IndexError, e:
			self.data = 'Invalid data: ' + str(ident[2])
		
		self.version = ident[3]
		
		self.os_abi_raw = ident[4]
		try:
			self.os_abi = self.os_abi_strings[ident[4]]
		except KeyError, e:
			self.os_abi = 'Invalid OS ABI: ' + str(ident[4])
		
		self.abi_version = ident[5]
		self.pad = ident[6]

	def __str__(self):
		return 'Magic: ' + self.magic + \
		       "\tClass: " + self.file_class + ' bit' +\
		       "\tData: " + self.data + "\n" + \
		       'Version: ' + str(self.version) + \
		       "\tOS ABI: " + self.os_abi + \
		       "\tABI Version: " + str(self.abi_version) + "\n";


# -----------------------------------------------------------------------------
class Ehdr(object):
	_format = "<16sHHIIIIIHHHHHH";
	type_strings = { 0: 'None', 
	                 1: 'Relocatable',
	                 2: 'Executable',
	                 3: 'Shared',
	                 4: 'Core' }
	
	machine_strings = { 0:'None', 
				1:'AT&T WE 32100', 
				2:'SUN SPARC',
				3:'Intel 80386', 
				4:'Motorola m68k family', 
				5:'Motorola m88k family', 
				7:'Intel 80860',
				8:'MIPS R3000 big-endian', 
				9:'IBM System/370', 
				10: 'MIPS R3000 little-endian',
				15: 'HPPA',
				17: 'Fujitsu VPP500',
				18: 'Sun "v8plus"',
				19: 'Intel 80960',
				20: 'PowerPC',
				21: 'PowerPC 64-bit',
				22: 'IBM S390',
			    36: 'NEC V800 series',
			    37: 'Fujitsu FR20',
			    38: 'TRW RH-32',
			    39: 'Motorola RCE',
			    40: 'ARM',
			    41: 'Digital Alpha',
			    42: 'Hitachi SH',
			    43: 'SPARC v9 64-bit',
			    44: 'Siemens Tricore',
			    45: 'Argonaut RISC Core',
			    46: 'Hitachi H8/300',
			    47: 'Hitachi H8/300H',
			    48: 'Hitachi H8S',
			    49: 'Hitachi H8/500',
			    50: 'Intel Merced`',
			    51: 'Stanford MIPS-X',
			    52: 'Motorola Coldfire',
			    53: 'Motorola M68HC12',
			    54: 'Fujitsu MMA Multimedia Accelerator',
			    55: 'Siemens PCP',
			    56: 'Sony nCPU embedded RISC',
			    57: 'Denso NDR1',
			    58: 'Motorola Start*Core',
			    59: 'Toyota ME16',
			    60: 'STMicroelectronic ST100',
			    61: 'Advanced Logic Corp. Tinyj',
			    62: 'AMD x86-64',
			    63: 'Sony DSP',
			    66: 'Siemens FX66',
			    67: 'STMicroelectronics ST9+ 8/16',
			    68: 'STmicroelectronics ST7',
			    69: 'Motorola MC68HC16',
			    70: 'Motorola MC68HC11',
			    71: 'Motorola MC68HC08',
			    72: 'Motorola MC68HC05',
			    73: 'Silicon Graphics SVx',
			    74: 'STMicroelectronics ST19',
			    75: 'Digital VAX',
			    76: 'Axis Communicationa',
			    77: 'Infineon Technologies',
			    78: 'Element 14 64-bit DSP',
			    79: 'LSI Logic 16-bit DSP',
			    80: 'Knuth MMIX',
			    81: 'Harvard University machine-independent obj',
			    82: 'SiTera Prism',
			    83: 'Atmel AVR',
			    84: 'Fujitsu FR30',
			    85: 'Mitsubishi D10V',
			    86: 'Mitsubishi D30V',
			    87: 'NEC v850',
			    88: 'Mitsubishi M32R',
			    89: 'Matsushita MN10300',
			    90: 'Matsushita MN10200',
			    91: 'picoJava',
			    92: 'OpenRISC',
			    93: 'ARC Cores Tangent-A5',
			    94: 'Tensilica Xtensa',
			    95: 'Num' }
	version_strings = ( 'None', 'Current' )

	def __init__(self, file, offset):
		handle = file.handle()
		size = struct.calcsize(self._format)
		try:
			handle.seek(offset)
			buf = handle.read(size)
		except Exception, e:
			sys.stderr.write( str(e) )
			raise e
		
		try:
			ehdr = struct.unpack(self._format, buf)
		except Exception, e:
			sys.stderr.write( str(e) )
			raise e
		
		self.ident = ehdr[0]
		
		self.type_raw = ehdr[1]
		try:
			self.type = self.type_strings[ehdr[1]]
		except KeyError, e:
			self.type = 'Invalid type: ' + str(ehdr[1])
		
		self.machine_raw = ehdr[2]
		try:
			self.machine = self.machine_strings[ehdr[2]]
		except KeyError, e:
			self.machine = 'Invalid machine: ' + str(ehdr[2])
		
		self.version_raw = ehdr[3]
		try:
			self.version = self.version_strings[ehdr[3]]
		except IndexError, e:
			self.version = 'Invalid version: ' + str(ehdr[3])
		
		self.entry = ehdr[4]
		self.phoff = ehdr[5]
		self.shoff = ehdr[6]
		self.flags = ehdr[7]
		self.ehsize = ehdr[8]
		self.phentsize = ehdr[9]
		self.phnum = ehdr[10]
		self.shentsize = ehdr[11]
		self.shnum = ehdr[12]
		self.shstrndx = ehdr[13]

	def __str__(self):
		return "Ident: " + self.ident + "\n" + \
			"Type: " + self.type + "\n" + \
			"Machine: " + self.machine + "\n" + \
			"Version: " + self.version + "\n" + \
			"Entry: " + hex(self.entry) + "\n" + \
			"PhOff: " + hex(self.phoff) + "\n" + \
			"ShOff: " + hex(self.shoff) + "\n" + \
			"Flags: " + hex(self.flags) + "\n" + \
			"EhSize: " + str(self.ehsize) + "\n" + \
			"PhEntSize: " + str(self.phentsize) + "\n" + \
			"PhNum: " + str(self.phnum) + "\n" + \
			"ShEntSize: " + str(self.shentsize) + "\n" + \
			"ShNum: " + str(self.shnum) + "\n" + \
			"ShStrNdx: " + str(self.shstrndx) + "\n"

class Ehdr32(Ehdr):
	_format = "<16sHHIIIIIHHHHHH";
	def __init__(self, file, offset):
		Ehdr.__init__(self, file, offset)

class Ehdr64(Ehdr):
	_format = "<16sHHIQQQIHHHHHH";
	def __init__(self, file, offset):
		Ehdr.__init__(self, file, offset)

# -----------------------------------------------------------------------------
class ElfSection(object):
	type_strings = { 0:'NULL', 1:'PROGBITS', 2:'SYMTAB', 3:'STRTAB', 
			 4:'RELA', 5:'HASH', 6:'DYNAMIC', 7:'NOTE', 
			 8:'NOBITS', 9:'REL', 10:'SHLIB', 11:'DYNSYM',
			 14:'INIT_ARRAY', 15:'FINI_ARRAY', 16:'PREINIT_ARRAY',
			 17:'GROUP', 18:'SYMTAB_SHNDX',
			 # GNU- and SUN-specific ones 
			 0x6ffffff7:'GNU_LIBLIST',
			 0x6ffffff8:'CHECKSUM',
			 0x6ffffffa:'SUNW_move',
			 0x6ffffffb:'SUNW_COMDAT',
			 0x6ffffffc:'SUNW_syminfo',
			 0x6ffffffd:'GNU_VERSION_DEF',
			 0x6ffffffe:'GNU_VERSION_NEED',
			 0x6fffffff:'GNU_VERSION_SYMTAB'};

	flag_strings = { 1:'WRITE', 2:'ALLOC', 4:'EXECINSTR', 8:'MERGE',
			16:'STRINGS', 32:'INFO_LINK', 64:'LINK_ORDER',
			128:'OS_NONCONFORMING', 256:'GROUP', 512:'TLS' }

	# default to 32-bit format
	_format = '<IIIIIIIIII';

	def __init__(self, file, offset):
		handle = file.handle()
		size = struct.calcsize(self._format)
		try:
			handle.seek(offset)
			buf = handle.read(size)
			if len(buf) < size:
				raise AssertionError, 'Buf size ' + \
				      str(len(buf)) + ' on read of ' + \
				      str(size) + "\n"
		except Exception, e:
			sys.stderr.write( str(e) )
			raise e
		
		try:
			shdr = struct.unpack(self._format, buf)
		except Exception, e:
			sys.stderr.write( str(e) )
			raise e
		
		self.name_raw = shdr[0]
		# self.name is determine when string table is known
		self.name = 'Unknown'
		
		self.type_raw = shdr[1]
		try:
			self.type = self.type_strings[shdr[1]]
		except KeyError, e:
			self.type = 'Invalid type: ' + str(shdr[1])
		
		self.flags_raw = shdr[2]
		self.flags = ''
		for i in self.flag_strings.iterkeys():
			if self.flags_raw & i:
				if len(self.flags):
					self.flags += '|'
				self.flags += self.flag_strings[i]
		
		self.addr = shdr[3]
		self.offset = shdr[4]
		self.size = shdr[5]
		self.link = shdr[6]
		self.info = shdr[7]
		self.addralign = shdr[8]
		self.entsize = shdr[9]


	def __str__(self):
		return  'Name: ' + self.name + \
		        "\tType: " + self.type + \
		        "\tFlags: " + self.flags + "\n" + \
		        'Offset: ' + hex(self.offset) + \
		        "\tAddress: " + hex(self.addr) + \
		        "\tSize: " + str(self.size) + "\n"

class ElfSection32(ElfSection):
	_format = '<IIIIIIIIII';
	def __init__(self, file, offset):
		ElfSection.__init__(self, file, offset)

class ElfSection64(ElfSection):
	_format = '<IIQQQQIIQI';
	def __init__(self, file, offset):
		ElfSection.__init__(self, file, offset)

# -----------------------------------------------------------------------------
class ElfSymbol(object):
	_format = '<IIIBBH'
	# bind = info >> 4, info = val & 0xf
	bind_strings = ( 'Local', 'Global', 'Weak' )
	type_strings = ( 'None', 'Object', 'Func', 'Section', 'File', 
	                 'Common', 'TLS' )
	# visibility = other & 0x03
	visibility_strings = ( 'Default', 'Internal', 'Hidden', 'Protected' )

class ElfSymbol32(ElfSymbol):
	_format = '<IIIBBH'

class ElfSymbol64(ElfSymbol):
	_format = '<IBBHQQ'
	
class ElfSymbolInfo(object):
	_format = '<HH'
	boundto_strings = { 0xFFFF : 'Self', 0xFFF3 : 'Parent' }
	flag_strings = { 1 : 'Direct', 2: 'PassThru', 4: 'Copy', 8: 'LazyLoad' }
	version_strings = ( 'None', 'Current' )


# -----------------------------------------------------------------------------
class ElfReloc(object):
	_format = '<II'

class ElfReloc32(ElfReloc):
	_format = '<II'
	# sym = info >> 8, type = info &  0xff

class ElfReloc64(ElfReloc):
	_format = '<QQ'
	# sym = info >> 32, type = info & 0xffffffff

class ElfRelocA32(ElfReloc32):
	_format = '<IIi'

class ElfRelocA64(ElfReloc64):
	_format = '<QQq'

# -----------------------------------------------------------------------------
class ElfSegment(object):
	_format = '<IIIIIIII'
	type_strings = ( 'Null', 'Load', 'Dynamic', 'Interp', 'Note', 'Shlib',
	                 'Phdr', 'TLS' )
	flag_strings = { 1 : 'X', 2 : 'W', 3 : 'R' }
	note_core_strings = { 1: 'PRStatus', 2:'FPRegSet', 3:'PRPSInfo',
			      4:'TaskStruct', 5:'Platform', 6:'AuxV',
			      7:'GWindows', 8:'ASRS', 10:'PStatus', 
			      13:'PSInfo', 14:'PRCred', 15:'UTSName',
			      16:'LWPStatus', 17:'LWPSInfo', 20:'PRFPXReg' }
	note_obj_strings = ( 'None', 'Version' )

	def __init__(self, file, offset):
		handle = file.handle()
		size = struct.calcsize(self._format)
		try:
			handle.seek(offset)
			buf = handle.read(size)
			if len(buf) < size:
				raise AssertionError, 'Buf size ' + \
				      str(len(buf)) + "\n"
		except Exception, e:
			sys.stderr.write( str(e) )
			raise e
		
		try:
			phdr = struct.unpack(self._format, buf)
		except Exception, e:
			sys.stderr.write( str(e) )
			raise e
		
		self.type_raw = phdr[0]
		try:
			self.type = self.type_strings[phdr[0]]
		except IndexError, e:
			self.type = 'Invalid type: ' + hex(phdr[0])
		
		self.offset = phdr[1]
		self.vaddr = phdr[2]
		self.paddr = phdr[3]
		self.file_size = phdr[4]
		self.mem_size = phdr[5]
		self.flags_raw = phdr[6]
		
		self.flags = ''
		for i in self.flag_strings.iterkeys():
			if self.flags_raw & i:
				if len(self.flags):
					self.flags += '|'
				self.flags += self.flag_strings[i]
		
		self.align = phdr[7]

	def __str__(self):
		return  'Type: ' + self.type + \
			"\tFlags: " + self.flags + "\n" + \
			'Offset: ' + hex(self.offset) + \
			"\tVirtual Address: " + hex(self.vaddr) + \
			"\tPhysical Address: " + hex(self.paddr) + "\n" + \
			"File Size: " + str(self.file_size) + \
			"\tMemory Size: " + str(self.mem_size) + \
			"\tAlignment: " + str(self.align) + "\n"

class ElfSegment32(ElfSegment):
	_format = '<IIIIIIII'
	def __init__(self, file, offset):
		ElfSegment.__init__(self, file, offset)

class ElfSegment64(ElfSegment):
	_format = '<IIQQQQQQ'
	def __init__(self, file, offset):
		handle = file.handle()
		size = struct.calcsize(self._format)
		try:
			handle.seek(offset)
			buf = handle.read(size)
			if len(buf) < size:
				raise AssertionError, 'Buf size ' + \
				      str(len(buf)) + "\n"
		except Exception, e:
			sys.stderr.write( str(e) )
			raise e
		
		try:
			phdr = struct.unpack(self._format, buf)
		except Exception, e:
			sys.stderr.write( str(e) )
			raise e
		
		self.type_raw = phdr[0]
		try:
			self.type = self.type_strings[phdr[0]]
		except IndexError, e:
			self.type = 'Invalid version: ' + hex(phdr[0])
		
		self.flags_raw = phdr[1]
		self.flags = ''
		for i in self.flag_strings.iterkeys():
			if self.flags_raw & i:
				if len(self.flags):
					self.flags += '|'
				self.flags += self.flag_strings[i]
		
		self.offset = phdr[2]
		self.vaddr = phdr[3]
		self.paddr = phdr[4]
		self.file_size = phdr[5]
		self.mem_size = phdr[6]
		self.align = phdr[7]

# -----------------------------------------------------------------------------
class ElfDynamic(object):
	_format = '<lI'

class ElfDynamic32(ElfDynamic):
	_format = '<lI'

class ElfDynamic64(ElfDynamic):
	_format = '<qQ'

# -----------------------------------------------------------------------------
class ELFFile(ObjFile.ObjFile):
	endian_strings = ( 'big endian', 'little endian', 'big endian' )
	endian_prefixes = ( '>', '<', '>' )

	cpu_strings = { 
		2:'sparc32',
		3:'ia32',
		4:'68k',
		20:'ppc32',
		21:'ppc64',
		40:'arm',
		41:'dec',
		43:'sparc64',
		50:'ia64',
		62:'x86-64',
	}

	def __init__(self, path=None, ident=None):
		super(ELFFile, self).__init__(path)
		self._format = 'ELF'
		self._epfx = '<'	# endian prefix
		self.parse()
		

	def get_string(self, offset):
		handle = self.handle()
		try:
			handle.seek(offset)
		except IndexError, e:
			raise e
		size = len(self) - offset
		if size > 256:
			size = 256
		buf = handle.read(size)
		if buf is None:
			return None
		regex = re.compile('([^\0]*\0)', re.DOTALL)
		match = regex.match(buf)
		if not match:
			return None
		name = match.groups(1)
		return name[0]


	def parse(self):
		# handle ELF ident
		handle = self.handle(0)
		self.elf_ident = EIdent(self, 0)
		
		# handle ELF header
		if self.elf_ident.file_class == '32' :
			self.elf_header = Ehdr32(self, 0)
			self.__word_size = 4
		elif self.elf_ident.file_class == '64' :
			self.elf_header = Ehdr64(self, 0)
			self.__word_size = 8
		else:
			# for now we default to 32-bit, maybe change to error
			# later...
			sys.stderr.write('Warning: Unknown ELF class ' + \
			                  self.elf_ident.file_class + "\n" )
			self.elf_header = Ehdr(self, 0)
			self.__word_size = 4
		
		# note: word size, endian, and cpu are internal
		# to ELFFile; they are passed on to the properties
		# of each appropriate sections.
		self.__endian = self.endian_strings[self.elf_ident.data_raw]
		self._epfx = self.endian_prefixes[self.elf_ident.data_raw]
		# set name of disassembler
		try:
			self._arch = \
				self.cpu_strings[self.elf_header.machine_raw]
		except KeyError, e:
			self._arch = 'unknown'
		
		# add elf entry point
		self._entry = self.elf_header.entry
		
		# NOTE: elf sections and segments are different from
		#       disasmbuf sections. all the elf stuff is created first,
		#       and the disasmbuf stuff is created based on that info.
		
		# handle section headers
		self.__section_headers()
		
		# handle program segment headers
		self.__program_headers()
		
		# handle sections
		sh_symtab = None
		sh_strtab = None
		sh_dynamic = None
		for sh in self.__section_list:
			if sh.type == 'SYMTAB':
				sh_symtab = sh
			elif sh.type == 'STRTAB':
				sh_strtab = sh
			elif sh.type == 'DYNAMIC':
				sh_dynamic = sh

		if sh_symtab and sh_strtab:
			self.handle_symtab(sh_symtab.offset, sh_symtab.size, 
		                   sh_strtab.offset, sh_strtab.size)

		# handle dynamic segment
		# TODO make this a method so we can avoid it in .o files
		format = '<II'
		size = struct.calcsize(format)
		offset = sh_dynamic.offset
		end_offset = offset + sh_dynamic.size
		tags = ('NULL', 'NEEDED', 'PLTRELSZ', 'PLTGOT', 'HASH',
			'STRTAB', 'SYMTAB', 'RELA', 'RELASZ', 'RELAENT',
			'STRSZ', 'SYMENT', 'INIT', 'FINI', 'SONAME', 'RPATH',
			'SYMBOLIC', 'REL', 'RELSZ', 'RELENT', 'PLTREL',
			'DEBUG', 'TEXTREL', 'JMPREL', 'BIND_NOW' )
		print 'DYN STRINGS SHDR: ' + hex(sh_dynamic.link)


		needed = []
		plt_size = None
		plt_addr = None
		dyn_strtab = None
		dyn_symtab = None
		rela_tbl = None
		rela_size = None
		rela_ent = None
		dyn_strtab_size = None
		dyn_symtab_ent = None
		init_addr = None
		fini_addr = None
		soname_off = None
		rpath_off = None
		rel_tbl = None
		rel_size = None
		rel_ent = None
		plt_type = None
		plt_reloc = None

		# NOTE: strtab is also in sh_link
		while offset < end_offset:
			print 'OFFSET ' + hex(offset) + ' : ' + hex(end_offset)
			handle.seek(offset)
			buf = handle.read(size)
			offset += size
			results = struct.unpack(format, buf)
			tag = results[0]
			val = results[1]
			try:
				name = tags[tag]
			except IndexError, e:
				print 'Invalid DYN TAG: ' + hex(tag)
				name = ''
			print 'TAG ' + hex(tag) + ' DYN ' + name + \
			      ' VAL ' + str(val)
			if tag == 0:
				break
			elif tag == 1:
				# val = stratb offset of library name
				needed.append(val)
			elif tag == 2:
				# val = total size of reloc entries in plt
				plt_size = val
			elif tag == 3:
				# address of plt and/or got
				plt_addr = val
			elif tag == 5:
				# address of strtab
				print "DYN STRTAB: " + hex(val)
				dyn_strtab = val
			elif tag == 6:
				# addr of symtab
				print "DYN SYMTAB: " + hex(val)
				dyn_symtab = val
			elif tag == 7:
				#addr of rela tbl
				rela_tbl = val
			elif tag == 8:
				# size of rela table
				rela_size = val
			elif tag == 9:
				# size of rela table entry
				rela_ent = val
			elif tag == 10:
				# size of strtab
				dyn_strtab_size = val
			elif tag == 11:
				# size of symtab entry
				dyn_symtab_ent = val
			elif tag == 12:
				# addr of init
				init_addr = val
			elif tag == 13:
				# addr of fini
				fini_addr = val
			elif tag == 14:
				# strtab offset of soname
				soname_off = val
			elif tag == 15:
				# stab offset of rpath
				rpath_off = val
			elif tag == 17:
				# addr of rel tbl
				rel_tbl = val
			elif tag == 18:
				# size of rel tbl
				rel_size = val
			elif tag == 19:
				# size of rel tbl entry
				rel_ent = val
			elif tag == 20:
				# val = type of plt entry. 7 = rela, 17 = rel
				plt_type = val
			#elif tag == 21:
			#elif tag == 22:
			elif tag == 23:
				# addr of reloc entries for plt
				plt_reloc = val
			#elif tag == 24:
		
		dyn_symtab_off = self.get_offset_for_rva(dyn_symtab)
		dyn_strtab_off = self.get_offset_for_rva(dyn_strtab)
		dyn_symtab_size = self.get_size_for_rva(dyn_symtab)

		self.handle_symtab( dyn_symtab_off, dyn_symtab_size,
		                    dyn_strtab_off, dyn_strtab_size )
		if soname_off:
			print 'SONAME : ' + self.get_string(dyn_strtab_off + \
			      soname_off)
		if rpath_off:
			print 'RPATH : ' + self.get_string(dyn_strtab_off + \
			      rpath_off)
		for n in needed:
			name = self.get_string(dyn_strtab_off + n)
			print 'NEEDED: ' + name


		# handle program segments
		for p in self.__segment_list:
			if p.type == 'Interp': 
				format = '<' + str(p.file_size) + 's'
				handle.seek(p.offset)
				buf = handle.read(p.file_size)
				results = struct.unpack(format, buf)
				print 'INTERP: ' + results[0]
		
		# create BigO sections
		self.__create_sections()

	def handle_symtab(self, symtab_off, symtab_sz, strtab_off, strtab_sz):
		# handle symtab
		bind_names = ( 'Local', 'Global', 'Weak' )
		type_names = ( 'None', 'Object', 'Function', 'Section',
		               'File' )
	
		handle = self.handle()
		format = '<LIIBBH'
		#format = 'IIIBBH'
		size = struct.calcsize(format)
		offset = symtab_off
		end_offset = offset + symtab_sz
		if end_offset > strtab_off:
			# this can happen in dynamic symbol tables where we
			# don't now the actual size, but just guess it.
			end_offset = strtab_off
		while offset < end_offset:
			handle.seek(offset)
			buf = handle.read(size)
			offset += size
			results = struct.unpack(format, buf)
			nm = results[0] # index in strtab
			value = results[1] # value of sym
			symsize = results[2] # size of sym
			info = results[3] # type/bind
			type = info & 0x0F
			bind = info >> 4
			other = results[4]  
			shndx = results[5] # section of symbol
			name = self.get_string(strtab_off + nm)

			if not name:
				name = ''
			try:
				type_str = type_names[type]
			except IndexError, e:
				print 'INVALID TYPE STR ' + hex(type)
				type_str = ''

			try:
				bind_str = bind_names[bind]
			except IndexError, e:
				print 'INVALID BIND STR ' + hex(bind)
				bind_str = ''

			line = 'SYM ' + str(nm) + ' ' + name +\
			       ' TYPE ' + type_str + \
			       ' BIND ' + bind_str + \
			       ' SIZE ' + str(symsize) + \
			       ' VAL ' + hex(value) + \
			       ' SEC ' + str(shndx)
			print 'SYM: ' + hex(offset) + ' : ' + line

	def get_offset_for_rva(self, rva):

		for s in self.__segment_list:
			if rva >= s.vaddr and rva < s.vaddr + s.file_size:
				diff = rva - s.vaddr
				return s.offset + diff
		return None

	def get_size_for_rva(self, rva):

		for s in self.__segment_list:
			if rva >= s.vaddr and rva < s.vaddr + s.file_size:
				diff = rva - s.vaddr
				return s.file_size - diff
		return None

	def __get_exec_segment(self):

		for ph in self.__segment_list:
			if ph.type == 'Load' and ph.flags.find('X') != -1:
				return ph
		return None

	def __section_shdr_details(self, sec, shdr):
			# TODO
		sec_type = { 'NULL' : sec.TYPE.FILEHDR, 
				'PROGBITS': sec.TYPE.PROGDATA, 
				'SYMTAB': sec.TYPE.SYMBOL, 
				'STRTAB': sec.TYPE.RESOURCE, 
				'RELA': sec.TYPE.RELOC,
				'HASH': sec.TYPE.FILEHDR,
				'DYNAMIC': sec.TYPE.FILEHDR,
				'NOTE': sec.TYPE.NOTE, 
				'NOBITS': sec.TYPE.PROGDATA,
				'REL': sec.TYPE.RELOC,
				'SHLIB': sec.TYPE.FILEHDR,
				'DYNSYM': sec.TYPE.IMPORT,
				'INIT_ARRAY': sec.TYPE.FILEHDR,
				'FINI_ARRAY': sec.TYPE.FILEHDR,
				'PREINIT_ARRAY': sec.TYPE.FILEHDR,
				'GROUP': sec.TYPE.FILEHDR,
				'SYMTAB_SHNDX': sec.TYPE.FILEHDR,
				'GNU_LIBLIST':sec.TYPE.FILEHDR,
				'CHECKSUM':sec.TYPE.FILEHDR,
				'SUNW_move':sec.TYPE.FILEHDR,
				'SUNW_COMDAT':sec.TYPE.FILEHDR,
				'SUNW_syminfo':sec.TYPE.FILEHDR,
				'GNU_VERSION_DEF':sec.TYPE.FILEHDR,
				'GNU_VERSION_NEED':sec.TYPE.FILEHDR,
				'GNU_VERSION_SYMTAB':sec.TYPE.FILEHDR,
				'.interp': sec.TYPE.PROGCODE,
				'.init':sec.TYPE.PROGCODE,
				'.fini':sec.TYPE.PROGCODE,
				'.text':sec.TYPE.PROGCODE,
				'.rodata':sec.TYPE.PROGDATA,
				'.data':sec.TYPE.PROGDATA,
				'.bss':sec.TYPE.PROGDATA,
				'.ctors':sec.TYPE.PROGCODE,
				'.dtors':sec.TYPE.PROGCODE };
		
		try:
			type = sec_type[shdr.name]
		except KeyError, e:
			type = None
		try:
			if type == None:
				type = sec_type[shdr.type]
		except KeyError, e:
			type = sec.TYPE.FILEHDR
		
		for f in shdr.flags.split('|'):
			if f == 'WRITE':
				sec._access |= sec.ACCESS.W
			elif f == 'ALLOC':
				sec._access |= sec.ACCESS.R
				sec._flags.append(sec.FLAGS.ALLOC)
			elif f == 'EXECINSTR':
				type = sec.TYPE.PROGCODE
				sec._access |= sec.ACCESS.X
				sec._arch = self._arch
			# TODO: determine if any of these are useful
			#elif f == 'MERGE':
			#elif f == 'STRINGS':
			#elif f == 'INFO_LINK':
			#elif f == 'LINK_ORDER':
			#elif f == 'OS_NONCONFORMING':
			#elif f == 'GROUP':
			#elif f == 'TLS':
		
		sec._type = type

	def __section_phdr_details(self, sec, phdr):
		sec_type = { 'Null' : sec.TYPE.FILEHDR, 
				'Load' : sec.TYPE.PROGDATA, 
				'Dynamic' : sec.TYPE.FILEHDR, 
				'Interp' : sec.TYPE.FILEHDR, 
				'Note' : sec.TYPE.NOTE, 
				'Shlib' : sec.TYPE.IMPORT,
				'Phdr' : sec.TYPE.FILEHDR, 
				'TLS' : sec.TYPE.FILEHDR }
		try:
			type = sec_type[phdr.type]
		except KeyError, e:
			type = sec.TYPE.FILEHDR
		
		if phdr.type == 'Load':
			sec._flags.append(sec.FLAGS.ALLOC)
		
		for f in phdr.flags.split('|'):
			if f == 'R':
				sec._access |= sec.ACCESS.R
			elif f == 'W':
				sec._access |= sec.ACCESS.W
			elif f == 'X':
				type = sec.TYPE.PROGCODE
				sec._access |= sec.ACCESS.X
				sec._arch = self._arch
		
		sec._type = type

	def __create_sections_from_shdr(self):
		sections = []
		
		self.__section_list.sort(lambda a,b: cmp(a.offset, b.offset))

		for sh in self.__section_list:
			if sh.flags.find('ALLOC') == -1: 
				# ignore sections that are not allocated
				continue
			# create new section
			s = ObjFile.Section(self, sh.name, sh.offset, 
					sh.size, sh.addr, self.__endian, 
					self.__word_size)
			# add section details
			self.__section_shdr_details(s, sh)
			# save in  file._sections
			self._sections.append( s )
			# add to local list 
			sections.append(s)
		
		sections.sort(lambda a, b: cmp(a.offset(), b.offset()))
		self.__segment_list.sort(lambda a,b: cmp(a.offset, b.offset))
		
		# create additional sections for unhandled, allocated
		# program segments.
		num = 1
		for ph in self.__segment_list:
			if ph.type != 'Load':
				# only interested in allocated segments
				continue
			
			# make a list of all sections in this segment
			sec = []
			for s in sections:
				if s.offset() >= ph.offset and \
					s.offset() < (ph.offset + ph.mem_size):
					sec.append(s)
			
			# if there are not any sections...
			if len(sec) < 1:
				# create section for entire segment
				new_s = ObjFile.Section( self, 
					'elf_' + str(num), ph.offset, 
					ph.mem_size, ph.vaddr,
					self.__endian, self.__word_size)
				self.__section_phdr_details(new_s, ph)
				self._sections.add(new_s)
				
				num += 1
				continue
			
			s = sec[0]
			if ph.offset < s.offset():
				# new section at start of segment
				if s.offset() < (ph.offset + ph.mem_size):
					size = s.offset() - ph.offset
				else :
					size = ph.mem_size
				
				new_s = ObjFile.Section( self, 
					'elf_' + str(num), ph.offset, 
					size, ph.vaddr, self.__endian, 
					self.__word_size)
				self.__section_phdr_details(new_s, ph)
				self._sections.append(new_s)
				
				num += 1
			
			last = sec[0]
			for s in sec[1:]:
				# if there is a gap between two segments
				diff = s.offset() - (last.offset()+len(last))
				if diff > 0:
					# create a section
					new_s = ObjFile.Section( self,
						'elf_' + str(num), 
						last.offset() + len(last), 
						diff, 
						last.rva() + \
						len(last),
						self.__endian, self.__word_size)
					self.__section_phdr_details(new_s, ph)
					self._sections.append(new_s)
					
					num += 1
					
				last = s
			
			# if there is unclaimed space at the end of the segment
			diff = (ph.offset + ph.mem_size) - \
			       (last.offset() + len(last))
			if diff > 0:
				# create a section
				new_s = ObjFile.Section( self, 
					'elf_' + str(num), 
						last.offset() + len(last),
						diff, 
						last.rva() + \
						len(last),
						self.__endian, self.__word_size)
				self.__section_phdr_details(new_s, ph)
				self._sections.append(new_s)
				num += 1


	def __create_sections(self):
		
		# check if there is a section header for code seg
		seg = self.__get_exec_segment()
		if seg == None:
			raise AssertionError, 'No executable program segment'
		
		for sh in self.__section_list:
			if sh.type == 'PROGBITS' and \
				sh.flags == 'ALLOC|EXECINSTR' and \
				sh.offset >= seg.offset and \
				(sh.offset + sh.size) <= \
				(seg.offset + seg.file_size):
				# elf section headers are sane: use them
				# to create the BGO sections
				self.__create_sections_from_shdr()
				return

		num = 1
		for ph in self.__segment_list:
			if ph.type == 'Load':
				s = ObjFile.Section( self, 'elf_' + str(num), 
					ph.offset, ph.mem_size, ph.vaddr,
					self.__endian, self.__word_size)
				self.__section_phdr_details(s, ph)
				self._sections.add(s)
				num += 1
			elif ph.type == 'Dynamic':
				# metadata -- handle specific dynamic types
				continue
			else:
				continue

	def __section_headers(self):
		
		self.__section_list = []
		for i in range( 0, self.elf_header.shnum):
			offset = self.elf_header.shoff + \
			         (i * self.elf_header.shentsize)
			print "SEC OFF %08X IN %08X" % (offset, self._size)
			if offset > self._size:
				raise AssertionError, "Section headers defined beyond EOF\n"
				break
			
			if self.elf_ident.file_class == '32' :
				sec = ElfSection32(self, offset)
			elif self.elf_ident.file_class == '64' :
				sec = ElfSection64(self, offset)
			else:
				# default to 32-bit
				sec = ElfSection(self, offset)
			
			self.__section_list.append(sec)
			
		# generate section names
		try:
			str_hdr = self.__section_list[self.elf_header.shstrndx]
		except IndexError, e:
			raise AssertionError,"Section strtab defined beyond EOF"
			return
		
		handle = self.handle()
		handle.seek(str_hdr.offset)
		strtab = handle.read(str_hdr.size) 
		for s in self.__section_list:
			try:
				names = strtab[s.name_raw:].split("\000")
				s.name =  names[0]
			except IndexError, e:
				sys.stderr.write('String extends beyond strtab')


	def __program_headers(self):
		handle = self.handle()
		self.__segment_list = []
		for i in range( 0, self.elf_header.phnum):
			offset = self.elf_header.phoff + \
			         (i * self.elf_header.phentsize)
			
			if self.elf_ident.file_class == '32' :
				sec = ElfSegment32(self, offset)
			elif self.elf_ident.file_class == '64' :
				sec = ElfSegment64(self, offset)
			else:
				# default to 32-bit
				sec = ElfSegment(self, offset)
			
			self.__segment_list.append(sec)

	def __str__(self):
		# MAGIC ident info
		buf = super(ELFFile, self).__str__() + "\n"
		# ELF header info
		buf += str(self.elf_ident) + "\n" + str(self.elf_header) + "\n"
		
		# program headers
		if len(self.__segment_list) > 0:
			buf += "\nProgram Segments:\n"
		for i in self.__segment_list:
			buf  += str(i) + "\n"
		
		# section headers
		if len(self.__section_list) > 0:
			buf += "\nSections:\n"
		for i in self.__section_list:
			buf  += str(i) + "\n"
		
		return buf


# -----------------------------------------------------------------------------
if __name__ == "__main__":
	if len(sys.argv) < 2:
		print "Usage: " + sys.argv[0] + " filename\n"
		sys.exit(1)

	# load and parse ELF file
	f = ELFFile( sys.argv[1], 'ELF 32' )

	print "__________________________________________________________\n"
	print "ELF File details for " + sys.argv[1] + ":\n"
	print str(f)

	# initialize disassembler
	native = Native.Syntax()
	insn_list = []
	raw_list = Ascii.List(Raw.Syntax(), insn_list)
	# uncomment this to test file generation
	#outfile = open("test.out", "w")
	#intel_file = Ascii.File(Intel.Syntax(), outfile)
	intel_stdout = Ascii.Stdout(Intel.Syntax())
	att_stdout = Ascii.Stdout(ATT.Syntax())

	# demonstrate different disassembly strategies
	if f.entry() is not None:
		print "___________________________________________________\n"
		print "Single disassembly at entry point:\n";
		try:
			offset = f.offset_for_rva(f.entry())
			dis = libdisasm.SingleDisassembler()
			insn = dis.disassemble(f, offset)
			print hex(offset) + ': ' + str(insn)
		except IndexError, e:
			print + str(e)

		print "__________________________________________________\n"
		print "CFlow disassembly from entry point:\n";
		try:
			offset = f.offset_for_rva(f.entry())
			#dis = libdisasm.ControlFlowDisassembler()
			dis = libdisasm.ControlFlowThreadedDisassembler()
			dis.disassemble(f, offset)
			for (rva, i) in f.instructions():
				i.output(intel_stdout)
		except IndexError, e:
			print + str(e)

	#print "__________________________________________________________\n"
	#print "CFlow disassembly from each symbol:\n";
	#dis = libdisasm.ControlFlowDisassembler()
	#for s in f.symbols():
		#dis.disassemble('cflow', buf)

	print "__________________________________________________________\n"
	print "Linear disassembly for each executable section:\n";

	dis = libdisasm.LinearDisassembler()
	for s in f.sections():
		if not s.access() & ObjFile.Section.ACCESS.X:
			continue
		print "Section " + s.name() + " (Offset: " + hex(s.offset()) +\
		      " Length: " + hex(len(s)) + "):"

		try:
			dis.disassemble(s)
		except (IndexError, EOFError), e:
			print "WARNING: Disassembly attempt to beyond end " +\
			      "of section " + s.name()

		for (rva, i) in s.instructions():
			line = '%08X: ' % i.rva() 
			count = 0
			for b in i.bytes():
				count += 1
				line += "%02X " % ord(b)
			for b in range( count, dis.max_insn_length() ):
				line += "   "
			line += str(i)
			print line

		print "\nTokens:\n"
		for (rva, i) in s.instructions():
			tokens = i.tokenize()
			line = ''
			for t in tokens:
				line += str(t) + ' '
			print line

		print "\nRaw Syntax:\n"

		# print using ASCII List Output object
		insn_list[:] = []	# clear list
		for (rva, i) in s.instructions():
			i.output(raw_list)
		for i in insn_list:
			print i + "\n"
		
		# print using Syntax object
		print "\nNative Syntax:\n"
		print native.header() + "\n"
		for (rva, i) in s.instructions():
			print native.instruction(i) + "\n"

		# print using ASCII STDOUT Output object
		print "\nATT Syntax:\n"
		for (rva, i) in s.instructions():
			i.output(att_stdout)

		# print using ASCII File Output object
		print "\nIntel Syntax:\n"
		for (rva, i) in s.instructions():
			#i.output(intel_file)
			i.output(intel_stdout)

